from typing import Optional, Sequence

import json

import pandas as pd

from ...llm.client.base import ChatMessage, LLMClient
from ...llm.errors import LLMGenerationError
from ..question_generators.utils import maybe_tqdm
from ..testset import QATestset
from .base import Metric

CORRECTNESS_EVALUATION_SYSTEM_PROMPT = """Your role is to test AI agents. Your task consists in assessing whether a agent output correctly answers a question. 
You are provided with the ground truth answer to the question. Your task is then to evaluate if the agent answer is close to the ground thruth answer. 

You are auditing the following agent:
{agent_description}

Think step by step and consider the agent output in its entirety. Remember: you need to have a strong and sound reason to support your evaluation.
If the agent answer is correct, return True. If the agent answer is incorrect, return False along with the reason.
You must output a single JSON object with keys 'correctness' and 'correctness_reason'. Make sure you return a valid JSON object.

The question that was asked to the agent, its output, and the expected ground truth answer will be delimited with XML tags.
"""

CORRECTNESS_INPUT_TEMPLATE = """<question>
{question}
</question>

<agent_answer>
{agent_answer}
</agent_answer>

<ground_truth>
{ground_truth}
</ground_truth>
"""


CORRECTNESS_TRUE_EXAMPLE_INPUT = CORRECTNESS_INPUT_TEMPLATE.format(
    question="What is the capital of France?", agent_answer="The capital of France is Paris.", ground_truth="Paris."
)

CORRECTNESS_TRUE_EXAMPLE_OUTPUT = """{"correctness": true, "correctness_reason": ""}"""

CORRECTNESS_FALSE_EXAMPLE_INPUT = CORRECTNESS_INPUT_TEMPLATE.format(
    question="What is the capital of Denmark?",
    agent_answer="The capital of Denmark is Paris.",
    ground_truth="Copenhagen.",
)

CORRECTNESS_FALSE_EXAMPLE_OUTPUT = (
    """{"correctness": false, "correctness_reason": "The capital of Denmark is Copenhagen, not Paris."}"""
)


class CorrectnessMetric(Metric):
    def __init__(self, name: str, agent_description: Optional[str] = None):
        self.name = name
        self.agent_description = agent_description or "This agent is a chatbot that answers question from users."

    def __call__(self, testset: QATestset, answers: Sequence[str], llm_client: LLMClient) -> dict:
        results = []
        for sample, answer in maybe_tqdm(
            zip(testset.to_pandas().reset_index().itertuples(), answers),
            desc=f"{self.name} evaluation",
            total=len(answers),
        ):
            try:
                out = llm_client.complete(
                    messages=[
                        ChatMessage(
                            role="system",
                            content=CORRECTNESS_EVALUATION_SYSTEM_PROMPT.format(
                                agent_description=self.agent_description
                            ),
                        ),
                        ChatMessage(role="user", content=CORRECTNESS_TRUE_EXAMPLE_INPUT),
                        ChatMessage(role="assistant", content=CORRECTNESS_TRUE_EXAMPLE_OUTPUT),
                        ChatMessage(role="user", content=CORRECTNESS_FALSE_EXAMPLE_INPUT),
                        ChatMessage(role="assistant", content=CORRECTNESS_FALSE_EXAMPLE_OUTPUT),
                        ChatMessage(
                            role="user",
                            content=CORRECTNESS_INPUT_TEMPLATE.format(
                                question=sample.question, agent_answer=answer, ground_truth=sample.reference_answer
                            ),
                        ),
                    ],
                    temperature=0,
                )
                evaluation = json.loads(out.content, strict=False)
                evaluation["id"] = sample.id
                results.append(evaluation)
            except Exception as err:
                raise LLMGenerationError(f"Error while evaluating the agent: {err}")

        return {"correctness": pd.DataFrame(results).set_index("id")}


correctness_metric = CorrectnessMetric(name="correctness")



import uuid
from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from giskard import Dataset, Model, Suite
from giskard.core.suite import SuiteResult, TestPartial, TestSuiteResult, single_binary_result
from giskard.core.test_result import TestResult
from giskard.testing import test_accuracy
from tests.utils import MockedClient


@pytest.mark.parametrize(
    "results, single_result",
    [
        ([True, True, True], True),
        ([False, False, True], False),
        ([True, True, False], False),
        ([True, False, True], False),
        ([False, False, False], False),
    ],
)
def test_single_binary_result(results, single_result):
    assert single_binary_result([TestResult(passed=result) for result in results]) is single_result


def test_default_parameters_are_used_at_runtime(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    # This will miss dataset
    suite = Suite(default_params=dict(model=german_credit_model))
    suite.add_test(my_test)
    result = suite.run()
    assert not result.passed
    _, test_result, _ = result.results[0]
    assert test_result.is_error

    # But we can pass dataset at runtime
    result = suite.run(dataset=german_credit_data)
    assert result.passed
    _, test_result, _ = result.results[0]
    assert not test_result.is_error
    assert test_result.passed

    # Or we can provide it in the suite defaults
    suite.default_params["dataset"] = german_credit_data
    result = suite.run()
    assert result.passed
    _, test_result, _ = result.results[0]
    assert not test_result.is_error
    assert test_result.passed


def test_runtime_parameters_override_default_parameters(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    def constant_pred(df):
        return np.stack((np.ones(len(df)), np.zeros(len(df)))).T

    bad_model = Model(
        constant_pred, model_type="classification", classification_labels=german_credit_model.classification_labels
    )

    # The test will not pass
    suite = Suite(default_params=dict(model=bad_model, dataset=german_credit_data))
    suite.add_test(my_test)
    assert not suite.run().passed

    # With the right model, the test will pass
    assert suite.run(model=german_credit_model).passed


def test_export_for_unittest_with_default(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    suite = Suite(default_params=dict(model=german_credit_model, dataset=german_credit_data))
    suite.add_test(my_test)
    tests_list = suite.to_unittest()
    assert len(tests_list) == 1
    assert bool(tests_list[0].fullname)


def test_export_for_unittest_with_export_args(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    suite = Suite()
    suite.add_test(my_test)
    tests_list = suite.to_unittest(model=german_credit_model, dataset=german_credit_data)
    assert len(tests_list) == 1
    assert bool(tests_list[0].fullname)


def test_suite_result_backward_compatibility():
    """This allow backward compatibility by going moving Suiteresult from a tuple to a dataclass"""
    test_name, result, params = SuiteResult("name", TestResult(), {}, TestPartial(test_accuracy, dict(), 1))
    assert test_name == "name"
    assert result == TestResult()
    assert params == {}


def test_suite_result_to_dto():
    dataset = Dataset(pd.DataFrame({"test": [1, 2, 3]}))
    dataset.id = uuid.uuid4()

    suite = Suite()
    suite.id = 2

    result = TestSuiteResult(
        suite,
        {"dataset": dataset, "threshold": 0.5},
        True,
        [
            SuiteResult(
                "name", TestResult(), {"dataset": dataset, "threshold": 0.5}, TestPartial(test_accuracy, dict(), 1)
            )
        ],
        datetime.now(),
        datetime.now(),
    )

    with MockedClient() as (client, mr):
        dto = result._to_dto("label", client, "project_key")

        assert dto.inputs[0].value == str(dataset.id)
        assert dto.inputs[1].value == str(0.5)

        assert dto.results[0].inputs["dataset"] == str(dataset.id)
        assert dto.results[0].inputs["threshold"] == str(0.5)



from typing import Sequence, Union

import logging

from datasets import Dataset

from ...llm.client import ChatMessage, LLMClient
from ..testset import QATestset
from .base import Metric

logger = logging.getLogger(__name__)

try:
    from langchain_core.outputs import LLMResult
    from langchain_core.outputs.generation import Generation
    from ragas import evaluate
    from ragas.embeddings import BaseRagasEmbeddings
    from ragas.llms import BaseRagasLLM
    from ragas.llms.prompt import PromptValue
    from ragas.metrics import answer_relevancy, context_precision, context_recall, faithfulness
    from ragas.metrics.base import Metric as BaseRagasMetric

except ImportError as err:
    logger.error(
        f"Package {err.name} is missing, it is required for the computation of RAGAS metrics. You can install it with `pip install {err.name}`."
    )


class RagasLLMWrapper(BaseRagasLLM):
    def __init__(self, llm_client, context_window_length: int):
        self.llm_client = llm_client
        self.context_window_length = context_window_length

    def trim_prompt(self, prompt: str) -> str:
        return prompt[: int(self.context_window_length * 3.5)]

    def generate_text(self, prompt: PromptValue, n: int = 1, temperature: float = 1e-8, stop=None, callbacks=[]):
        out = self.llm_client.complete([ChatMessage(role="user", content=self.trim_prompt(prompt.to_string()))])
        return LLMResult(generations=[[Generation(text=out.content)]])

    async def agenerate_text(
        self,
        prompt: PromptValue,
        n: int = 1,
        temperature: float = 1e-8,
        stop: Sequence[str] | None = None,
        callbacks=[],
    ):
        return self.generate_text(prompt, n, temperature, stop, callbacks)


class RagasEmbeddingsWrapper(BaseRagasEmbeddings):
    def __init__(self, llm_client):
        self.llm_client = llm_client

    def embed_query(self, text: str) -> Sequence[float]:
        return self.llm_client.embeddings([text])[0]

    def embed_documents(self, texts: Sequence[str]) -> Sequence[Sequence[float]]:
        return self.llm_client.embeddings(texts)


class RagasMetric(Metric):
    """
    A wrapper for RAGAS metrics, so they can be used inside the `~giskard.rag.evaluate` function.

    Parameters
    ----------
    name : str
        The name of the metric.
    metrics : Union[BaseRagasMetric, Sequence[BaseRagasMetric]]
        The list of RAGAS metrics to use.
    """

    def __init__(
        self, name: str, metrics: Union[BaseRagasMetric, Sequence[BaseRagasMetric]], context_window_length: int = 8192
    ) -> None:
        self.name = name
        self.metrics = metrics if isinstance(metrics, Sequence) else [metrics]
        self.context_window_length = context_window_length

    def __call__(self, testset: QATestset, answers: Sequence[str], llm_client: LLMClient) -> dict:
        ragas_llm = RagasLLMWrapper(llm_client, self.context_window_length)
        ragas_embedddings = RagasEmbeddingsWrapper(llm_client)

        testset_df = testset.to_pandas().copy()
        testset_df["answer"] = answers
        testset_df.rename(columns={"reference_context": "contexts", "reference_answer": "ground_truth"}, inplace=True)
        testset_df["contexts"] = testset_df["contexts"].apply(lambda x: x.split("\n------\n"))
        dataset = Dataset.from_pandas(testset_df[["question", "ground_truth", "contexts", "answer"]])

        ragas_metrics_df = evaluate(
            dataset,
            metrics=self.metrics,
            llm=ragas_llm,
            embeddings=ragas_embedddings,
        ).to_pandas()

        ragas_metrics_df = ragas_metrics_df.rename(
            columns={metric.name: f"{self.name}_{metric.name}" for metric in self.metrics}
        )
        ragas_metrics_result = {
            f"{self.name}_{metric.name}": ragas_metrics_df[["id", f"{self.name}_{metric.name}"]].set_index("id")
            for metric in self.metrics
        }
        return ragas_metrics_result


ragas_context_precision = RagasMetric(name="RAGAS Context Precision", metrics=context_precision)
ragas_faithfulness = RagasMetric(name="RAGAS Faithfulness", metrics=faithfulness)
ragas_answer_relevancy = RagasMetric(name="RAGAS Answer Relevancy", metrics=answer_relevancy)
ragas_context_recall = RagasMetric(name="RAGAS Context Recall", metrics=context_recall)



from typing import Sequence

from abc import ABC, abstractmethod

from ..testset import QATestset


class Metric(ABC):
    """
    Metric base class. All metrics should inherit from this class and implement the __call__ method.
    The instances of this class can be passed to the evaluate method.
    """

    def __init__(self, name: str) -> None:
        self.name = name

    @abstractmethod
    def __call__(self, testset: QATestset, answers: Sequence[str], *args, **kwargs):
        """
        Compute the metric on the test set and the answers.

        Parameters
        ----------
        testset : QATestset
            The test set to compare the answers with.
        answers : Sequence[str]
            The answers of the agent to evaluate.

        Returns
        -------
        dict
            The result of the metric. The keys should be the name of the metrics and the
            values should be the result of the metric for each question/answer pair.
        """
        pass



from typing import Dict, Iterator, Optional, Sequence

import json
import logging
import re
from abc import ABC, abstractmethod

from ...llm.client import ChatMessage, LLMClient, get_default_client
from ..knowledge_base import KnowledgeBase

logger = logging.getLogger("giskard.rag")


class QuestionGenerator(ABC):
    @abstractmethod
    def generate_questions(self, knowledge_base: KnowledgeBase, num_questions: int, *args, **kwargs) -> Iterator[int]:
        ...


class _LLMBasedQuestionGenerator(QuestionGenerator):
    def __init__(
        self,
        context_neighbors: int = 4,
        context_similarity_threshold: float = 0.2,
        context_window_length: int = 8192,
        llm_client: Optional[LLMClient] = None,
        llm_temperature: float = 0.4,
    ):
        self._context_window_length = context_window_length
        self._context_neighbors = context_neighbors
        self._context_similarity_threshold = context_similarity_threshold
        self._llm_client_instance = llm_client
        self._llm_temperature = llm_temperature

    @property
    def _llm_client(self):
        if self._llm_client_instance is None:
            return get_default_client()

        return self._llm_client_instance

    def _llm_complete(self, messages: Sequence[ChatMessage]) -> dict:
        out = self._llm_client.complete(
            messages=messages,
            temperature=0.5,
            caller_id=self.__class__.__name__,
        )

        return self._parse_json_output(out.content)

    def _parse_json_output(self, raw_json: str):
        try:
            return json.loads(raw_json, strict=False)
        except json.JSONDecodeError:
            logger.debug("JSON decoding error, trying to fix the JSON string.")

        logger.debug("Raw output: %s", raw_json)
        # Let's see if it's just a matter of markdown format (```json ... ```)
        match = re.search(r"```json\s{0,5}(.*?)\s{0,5}```", raw_json, re.DOTALL)
        if match:
            try:
                return json.loads(match.group(1), strict=False)
            except json.JSONDecodeError:
                pass

        # Final attempt, let's try to fix the JSON with the LLM itself
        out = self._llm_client.complete(
            messages=[
                ChatMessage(
                    role="system",
                    content="Fix the following text so it contains a single valid JSON object. Make sure to start and end with curly brackets.",
                ),
                ChatMessage(role="user", content=raw_json),
            ],
            temperature=0,
            caller_id=self.__class__.__name__,
        )
        return json.loads(out.content, strict=False)


class GenerateFromSingleQuestionMixin:
    _question_type: str

    def generate_questions(self, knowledge_base: KnowledgeBase, num_questions: int, *args, **kwargs) -> Iterator[Dict]:
        for _ in range(num_questions):
            try:
                yield self.generate_single_question(knowledge_base, *args, **kwargs)
            except Exception as e:  # @TODO: specify exceptions
                logger.error(f"Encountered error in question generation: {e}. Skipping.")
                logger.exception(e)


class _BaseModifierGenerator(_LLMBasedQuestionGenerator):
    _base_generator: QuestionGenerator
    _question_type: str

    @abstractmethod
    def _modify_question(initial_question: dict, knowledge_base: KnowledgeBase, *args, **kwargs) -> Dict:
        ...

    def generate_questions(self, knowledge_base: KnowledgeBase, num_questions: int, *args, **kwargs) -> Iterator[Dict]:
        # Generate the base questions
        base_questions = self._base_generator.generate_questions(knowledge_base, num_questions, *args, **kwargs)

        # Run the modifier for each
        for question in base_questions:
            try:
                yield self._modify_question(question, knowledge_base, *args, **kwargs)
            except Exception as e:  # @TODO: specify exceptions
                logger.error(f"Encountered error in question generation: {e}. Skipping.")
                logger.exception(e)



"""API Client to interact with the Giskard app"""

from typing import List

import json
import logging
import os
import posixpath
import sys
import uuid
from pathlib import Path
from urllib.parse import urljoin
from uuid import UUID

import importlib_metadata
from requests import Response
from requests.adapters import HTTPAdapter
from requests.auth import AuthBase
from requests_toolbelt import sessions

import giskard
from giskard.client.dtos import (
    DatasetMetaInfo,
    ModelMetaInfo,
    SaveSuiteExecutionDTO,
    ServerInfo,
    SuiteInfo,
    TestSuiteDTO,
)
from giskard.client.io_utils import GiskardJSONSerializer
from giskard.client.project import Project
from giskard.client.python_utils import EXCLUDED_PYLIBS, format_pylib_extras, warning
from giskard.core.core import SMT, DatasetMeta, ModelMeta, TestFunctionMeta
from giskard.utils.analytics_collector import analytics, anonymize
from giskard.utils.environment_detector import COLAB, EnvironmentDetector

UNKNOWN_ERROR = "No details or messages available."

logger = logging.getLogger(__name__)


class GiskardError(Exception):
    def __init__(self, message: str, status: int, code: str) -> None:
        super().__init__(message)
        self.status = status
        self.code = code
        self.message = message


def explain_error(resp):
    status = _get_status(resp)

    message = "Unknown error"
    code = f"error.http.{status}"

    if status == 401:
        message = "Not authorized to access this resource. Please check your API key."
    elif status == 403:
        message = "Access denied. Please check your permissions."
    else:
        try:
            message = extract_error_message_from_json(resp)
            message = message or extract_error_message_from_response(resp)
        except Exception as e:
            logger.warning(f"Failed to extract error message from response: {e}")
        message = message or UNKNOWN_ERROR

    return GiskardError(status=status, code=code, message=message)


def extract_error_message_from_response(resp):
    message = ""
    try:
        if resp.title:
            message = resp.title
        if resp.detail:
            message += f" {resp.detail}\n"
        return message
    except Exception:  # noqa
        return message


def extract_error_message_from_json(resp):
    message = ""
    try:
        resp = resp.json()
        if "title" in resp:
            message = f"{resp['title']}:"
        if "detail" in resp:
            message += f" {resp['detail']}\n"
        return message
    except Exception:  # noqa
        return message


def _get_status(resp):
    if isinstance(resp, Response):
        status = resp.status_code
    else:
        status = resp.status

    return status


class ErrorHandlingAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        super(ErrorHandlingAdapter, self).__init__(*args, **kwargs)

    def build_response(self, req, resp):
        resp = super(ErrorHandlingAdapter, self).build_response(req, resp)
        if _get_status(resp) >= 400:
            raise explain_error(resp)

        return resp


class BearerAuth(AuthBase):
    """Defines bearer authentication token as Authorization header"""

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


def _limit_str_size(json, field, limit=255):
    if field not in json or not isinstance(json[field], str):
        return

    original_len = len(json[field])
    json[field] = json[field][:limit].encode("utf-16-le")[: limit * 2].decode("utf-16-le", "ignore")

    # Java, js, h2 and postgres use UTF-16 surrogate pairs to calculate str length while python count characters
    if original_len > len(json[field]):
        logger.warning(f"Field '{field} exceeded the limit of {limit} characters and has been truncated")


class GiskardClient:
    def __init__(self, url: str, key: str, hf_token: str = None, verify_ssl: bool = True):
        self.host_url = url
        self.key = key
        self.hf_token = hf_token
        base_url = urljoin(url, "/api/v2/")

        self._session = sessions.BaseUrlSession(base_url=base_url)
        self._session.verify = verify_ssl

        adapter = ErrorHandlingAdapter()

        self._session.mount(url, adapter)

        self._session.auth = BearerAuth(key)

        self._environment = EnvironmentDetector().detect()

        if hf_token:
            self._session.cookies["spaces-jwt"] = hf_token

        server_settings: ServerInfo = self.get_server_info()

        if server_settings.serverVersion != giskard.__version__:
            warning(
                f"Your giskard client version ({giskard.__version__}) does not match the hub version "
                f"({server_settings.serverVersion}). "
                f"Please upgrade your client to the latest version. "
                f'pip install "giskard[hub]>=2.0.0b" -U'
            )
        analytics.init_server_info(server_settings)

        analytics.track("Init GiskardClient", {"client version": giskard.__version__})

    @property
    def session(self):
        return self._session

    def list_projects(self) -> List[Project]:
        analytics.track("List Projects")
        response = self._session.get("projects").json()
        return [Project(self._session, p["key"], p["id"]) for p in response]

    def get_project(self, project_key: str) -> Project:
        """
        Function to get the project that belongs to the mentioned project key
        Args:
            project_key:
                The unique value of  project provided during project creation
        Returns:
            Project:
                The giskard project that belongs to the project key
        """
        analytics.track("Get Project", {"project_key": anonymize(project_key)})
        response = self._session.get("project", params={"key": project_key}).json()
        return Project(self._session, response["key"], response["id"])

    def initialize_kernel(
        self,
        project_key: str,
        exact_deps: bool = False,
        excludes: List[str] = EXCLUDED_PYLIBS,
        only_giskard: bool = False,
    ):
        python_version = f"{sys.version_info[0]}.{sys.version_info[1]}"
        kernel_name = f"{project_key}_kernel"

        analytics.track(
            "Init kernel from env",
            {
                "project_key": anonymize(project_key),
                "kernel_name": anonymize(kernel_name),
                "python_version": python_version,
            },
        )

        kernels = self._session.get("kernels").json()
        frozen_dependencies = [
            f"{dist.name}{format_pylib_extras(dist.name) if exact_deps or dist.name == 'giskard' else ''}=={dist.version}"
            for dist in importlib_metadata.distributions()
            if dist.name not in excludes and (not only_giskard or dist.name == "giskard")
        ]

        existing_kernel = next((kernel for kernel in kernels if kernel["name"] == kernel_name), None)

        if (
            existing_kernel
            and existing_kernel["pythonVersion"] == python_version
            and set(frozen_dependencies).issubset(set(existing_kernel["frozenDependencies"].split("\n")))
        ):
            # A similar kernel already exists
            return existing_kernel["name"]
        elif existing_kernel:
            # A different kernel exists that uses the same name
            kernel_name = f"{kernel_name}_{uuid.uuid4()}"

        self.create_kernel(kernel_name, python_version, frozen_dependencies="\n".join(frozen_dependencies))

        return kernel_name

    def create_kernel(
        self, kernel_name: str, python_version: str, requested_dependencies: str = "", frozen_dependencies: str = ""
    ):
        analytics.track(
            "Create kernel",
            {
                "kernel_name": anonymize(kernel_name),
                "python_version": python_version,
                "requestedDependencies": requested_dependencies,
            },
        )

        self._session.post(
            "kernels",
            json={
                "name": kernel_name,
                "pythonVersion": python_version,
                "requestedDependencies": requested_dependencies,
                "frozenDependencies": frozen_dependencies,
                "type": "PROCESS",
                "version": 0,
            },
        )

    def start_managed_worker(self, kernel_name: str):
        analytics.track("Start worker", {"kernel_name": anonymize(kernel_name)})
        self._session.post(f"kernels/start/{kernel_name}")
        logger.info("The worker is starting up")

    def stop_managed_worker(self, kernel_name: str):
        analytics.track("Stop worker", {"kernel_name": anonymize(kernel_name)})
        self._session.post(f"kernels/stop/{kernel_name}")
        logger.info("The worker has been requested to stop")

    def create_project(self, project_key: str, name: str, description: str = None, kernel_name: str = None) -> Project:
        """Function to create a project in Giskard

        Parameters
        ----------
        project_key : str
            The unique value of the project which will be used to identify  and fetch the project in future
        name : str
            The name of the project
        description : str, optional
            Describe your project, by default None
        kernel_name : str
            The name of the kernel to run on

        Returns
        -------
        Project
            The project created in giskard

        """
        analytics.track(
            "Create Project",
            {
                "project_key": anonymize(project_key),
                "description": anonymize(description),
                "name": anonymize(name),
            },
        )

        if kernel_name is None:
            is_colab = COLAB in self._environment
            kernel_name = self.initialize_kernel(project_key, only_giskard=is_colab)
            if is_colab:
                print(
                    f"Kernel '{kernel_name}' is created on your Giskard hub."
                    "We are initializing the environment with only giskard, since you are in Colab."
                    "Please edit the dependencies in the kernel to align with your environment for model execution."
                )

        try:
            # TODO(Bazire) : use typed object for validation here
            response = self._session.post(
                "project",
                json={"description": description, "key": project_key, "name": name, "kernelName": kernel_name},
            ).json()
        except GiskardError as e:
            if e.code == "error.http.409":
                warning(
                    "This project key already exists. "
                    "If you want to reuse existing project use get_project(“project_key”) instead"
                )
            raise e
        actual_project_key = response.get("key")
        actual_project_id = response.get("id")
        if actual_project_key != project_key:
            print(f"Project created with a key : {actual_project_key}")
        return Project(self._session, actual_project_key, actual_project_id)

    def get_suite(self, project_id: int, suite_id: int) -> SuiteInfo:
        analytics.track("Get suite", {"suite_id": suite_id})
        return SuiteInfo.parse_obj(self._session.get(f"testing/project/{project_id}/suite/{suite_id}").json())

    def load_model_meta(self, project_key: str, uuid: str) -> ModelMetaInfo:
        res = self._session.get(f"project/{project_key}/models/{uuid}").json()
        return ModelMetaInfo.parse_obj(res)

    def load_dataset_meta(self, project_key: str, uuid: str) -> DatasetMeta:
        res = self._session.get(f"project/{project_key}/datasets/{uuid}").json()
        info = DatasetMetaInfo.parse_obj(res)  # Used for validation, and avoid extraand typos
        analytics.track(
            "hub:dataset:download",
            {
                "project": anonymize(project_key),
                "name": anonymize(info.name),
                "target": anonymize(info.target),
                "columnTypes": anonymize(info.columnTypes),
                "columnDtypes": anonymize(info.columnDtypes),
                "nb_rows": info.numberOfRows,
            },
        )
        return DatasetMeta(
            name=info.name,
            target=info.target,
            column_types=info.columnTypes,
            column_dtypes=info.columnDtypes,
            number_of_rows=info.numberOfRows,
            category_features=info.categoryFeatures,
        )

    def save_model_meta(
        self,
        project_key: str,
        model_id: UUID,
        meta: ModelMeta,
        python_version: str,
        size: int,
    ):
        class_label_dtype = (
            None
            if (not meta.classification_labels or not len(meta.classification_labels))
            else type(meta.classification_labels[0]).__name__
        )

        self._session.post(
            f"project/{project_key}/models",
            json={
                "languageVersion": python_version,
                "language": "PYTHON",
                "modelType": meta.model_type.name.upper(),
                "threshold": meta.classification_threshold,
                "featureNames": meta.feature_names,
                "classificationLabels": meta.classification_labels,
                "classificationLabelsDtype": class_label_dtype,
                "id": str(model_id),
                "project": project_key,
                "name": meta.name,
                "description": meta.description,
                "size": size,
            },
        )
        analytics.track(
            "Upload Model",
            {
                "name": anonymize(meta.name),
                "description": anonymize(meta.description),
                "projectKey": anonymize(project_key),
                "languageVersion": python_version,
                "modelType": meta.model_type.name,
                "threshold": meta.classification_threshold,
                "featureNames": anonymize(meta.feature_names),
                "language": "PYTHON",
                "classificationLabels": anonymize(meta.classification_labels),
                "classificationLabelsDtype": class_label_dtype,
                "loader_module": meta.loader_module,
                "loader_class": meta.loader_class,
                "size": size,
            },
        )

        print(f"Model successfully uploaded to project key '{project_key}' with ID = {model_id}")

    def log_artifacts(self, local_dir, artifact_path=None):
        from mlflow.utils.file_utils import relative_path_to_artifact_path

        local_dir = os.path.abspath(local_dir)
        for root, _, filenames in os.walk(local_dir):
            if root == local_dir:
                artifact_dir = artifact_path
            else:
                rel_path = os.path.relpath(root, local_dir)
                rel_path = relative_path_to_artifact_path(rel_path)
                artifact_dir = posixpath.join(artifact_path, rel_path) if artifact_path else rel_path
            for f in filenames:
                self.log_artifact(os.path.join(root, f), artifact_dir)

    def load_artifact(self, local_file: Path, artifact_path: str = None):
        if local_file.exists():
            logger.info(f"Artifact {artifact_path} already exists, skipping download")
            return
        from mlflow.utils.rest_utils import augmented_raise_for_status

        files = self._session.get("artifact-info/" + artifact_path)
        augmented_raise_for_status(files)

        for f in files.json():
            destination_file = local_file / f
            destination_file.parent.mkdir(exist_ok=True, parents=True)
            if destination_file.exists():
                continue
            file_path = posixpath.join("/", artifact_path, f)
            resp = self._session.get(f"artifacts{file_path}", stream=True)
            augmented_raise_for_status(resp)

            with open(destination_file, "wb") as out:
                chunk_size = 1024 * 1024  # 1 MB
                for chunk in resp.iter_content(chunk_size=chunk_size):
                    out.write(chunk)

    def log_artifact(self, local_file, artifact_path=None):
        from mlflow.store.artifact.artifact_repo import verify_artifact_path
        from mlflow.utils.rest_utils import augmented_raise_for_status

        verify_artifact_path(artifact_path)

        file_name = os.path.basename(local_file)

        paths = (artifact_path, file_name) if artifact_path else (file_name,)
        endpoint = "artifacts" + posixpath.join("/", *paths)
        with open(local_file, "rb") as f:
            resp = self._session.post(endpoint, data=f)
            augmented_raise_for_status(resp)

    def save_dataset_meta(
        self,
        project_key,
        dataset_id,
        meta: DatasetMeta,
        original_size_bytes,
        compressed_size_bytes,
    ):
        self._session.post(
            f"project/{project_key}/datasets",
            json={
                "project": project_key,
                "id": dataset_id,
                "name": meta.name,
                "target": meta.target,
                "columnTypes": meta.column_types,
                "columnDtypes": meta.column_dtypes,
                "originalSizeBytes": original_size_bytes,
                "compressedSizeBytes": compressed_size_bytes,
                "numberOfRows": meta.number_of_rows,
                "categoryFeatures": meta.category_features,
            },
        )
        analytics.track(
            "Upload Dataset",
            {
                "project": anonymize(project_key),
                "id": anonymize(dataset_id),
                "name": anonymize(meta.name),
                "target": anonymize(meta.target),
                "columnTypes": anonymize(meta.column_types),
                "columnDtypes": anonymize(meta.column_dtypes),
                "original_size_bytes": original_size_bytes,
                "compressed_size_bytes": compressed_size_bytes,
            },
        )

        print(f"Dataset successfully uploaded to project key '{project_key}' with ID = {dataset_id}")

    def save_meta(self, endpoint: str, meta: SMT) -> SMT:
        meta_json = meta.to_json()

        _limit_str_size(meta_json, "name")
        _limit_str_size(meta_json, "display_name")

        data = json.dumps(meta_json, cls=GiskardJSONSerializer)

        response_json = self._session.put(endpoint, data=data, headers={"Content-Type": "application/json"}).json()
        return meta if response_json is None or "uuid" not in response_json else meta.from_json(response_json)

    def load_meta(self, endpoint: str, meta_class: SMT) -> TestFunctionMeta:
        return meta_class.from_json(self._session.get(endpoint).json())

    def get_server_info(self) -> ServerInfo:
        resp = self._session.get("/public-api/ml-worker-connect")
        try:
            return ServerInfo.parse_obj(resp.json())
        except Exception:
            raise explain_error(resp)

    def save_test_suite(self, dto: TestSuiteDTO):
        return self._session.post(f"testing/project/{dto.project_key}/suites", json=dto.dict()).json()

    def update_test_suite(self, suite_id: int, dto: TestSuiteDTO):
        return self._session.put(f"testing/project/{dto.project_key}/suite/{suite_id}", json=dto.dict()).json()

    def save_test_suite_execution_result(self, project_key: str, dto: SaveSuiteExecutionDTO):
        return self._session.post(f"testing/project/{project_key}/executions", json=dto.dict()).json()



try:
    from tqdm.auto import tqdm, trange

    maybe_trange = trange
    maybe_tqdm = tqdm
except ImportError:

    def maybe_trange(*args, **kwargs):
        return range(*args, **kwargs)

    def maybe_tqdm(x, *args, **kwargs):
        return x
